#include "cuda_runtime.h"
#include "device_launch_parameters.h"
#include <device_functions.h>

#include <iostream>
#include <cstdlib>
#include <ctime>
#include <stdio.h>
#include <string>

#include <opencv2\core.hpp>
#include <opencv2\opencv_modules.hpp>
#include <opencv2\opencv.hpp>

using namespace cv;

#define BRUSH_SIZE 5
#define LEVELS 20

__global__ void oilPaint(uchar3* const src, uchar3* const dst, const int width, const int height) {
	int CountIntensity[LEVELS];
	int RedAverage[LEVELS];
	int GreenAverage[LEVELS];
	int BlueAverage[LEVELS];

	for (int k = 0; k < LEVELS; k++) {
		CountIntensity[k] = 0;
		RedAverage[k] = 0;
		GreenAverage[k] = 0;
		BlueAverage[k] = 0;
	}

	//2D Index of current thread
	const int x = blockIdx.x * blockDim.x + threadIdx.x;
	const int y = blockIdx.y * blockDim.y + threadIdx.y;


	//check if index is out of bounds 
	if ((x < width) && (y < height)) {
		for (int i = -BRUSH_SIZE; i <= BRUSH_SIZE; ++i) {//all surrounding above and below pixels 		
			int crtY = y + i; // neighbour pixel index
			if (crtY >= 0 && crtY < height) {
				for (int j = -BRUSH_SIZE; j <= BRUSH_SIZE; ++j) {//all surrounding right/left pixels 		
					int crtX = x + j;// neighbour pixel index
					if (crtX >= 0 && crtX < width) { //check if neighbour pixel is out of bounds
						const uchar3 curPix = src[crtY  * width + crtX];
						
						//finds average of all colour values r b and g 
						//change that value to an intensity percentage 
						//change that percentage to index (0 - LEVELS is the range)
						//increases intensity count at the index that repersents that level 						
						int tmp = (((curPix.x + curPix.y + curPix.z) / 3.0) * LEVELS) / 255.0;;
						
						//adds current pixels value to the array at index that repersents that level
						CountIntensity[tmp]++;
						
						//adds current pixels value to the array at index that repersents that level
						BlueAverage[tmp] += curPix.x;
						GreenAverage[tmp] += curPix.y;
						RedAverage[tmp] += curPix.z;
					}
				}
			}
		}

		int nCurMax = 0;
		int nMaxIndex = 0;
		
		for (int nI = 0; nI < LEVELS; nI++) {
			if (CountIntensity[nI] > nCurMax) {
				nCurMax = CountIntensity[nI];
				nMaxIndex = nI;
			}
		}

		// check for 0
		if (nCurMax == 0) { nCurMax = 1; }

		uchar3 color;
		color.x = BlueAverage[nMaxIndex] / nCurMax;;
		color.y = GreenAverage[nMaxIndex] / nCurMax;
		color.z = RedAverage[nMaxIndex] / nCurMax;

		dst[y * width + x] = color;
	}
}

int main(int argc, char** argv) {
	//check command line arguments 
	if (argc != 2) {
		std::cout << argv[0] << ": invalid number of arguments\n";
		return 1;
	}
	std::string fileName = argv[1];

	//read picture 
	Mat src = imread(fileName);
	if (src.empty()) {
		std::cerr << "File not found/read.";
		std::cin.get();
		return -1;
	}

	//define sizes
	const size_t height = src.rows;
	const size_t width = src.cols;

	//print picutre 
	imshow("Source image", src);

	//device setup 
	int deviceNum;
	cudaDeviceProp prop;
	cudaGetDevice(&deviceNum);
	cudaGetDeviceProperties(&prop, deviceNum);
	int ntpb =  floor(sqrt( prop.maxThreadsDim[0]));
	int nblks = (width*height + ntpb - 1) / ntpb;

	//allocate memory 
	uchar3 *d_src, *d_dst;
	cudaMalloc<uchar3>(&d_src, height * width * sizeof(uchar3));
	cudaMalloc<uchar3>(&d_dst, height * width * sizeof(uchar3));

	//copy to device
	cudaMemcpy(d_src, src.ptr<uchar3>(), height * width * sizeof(uchar3), cudaMemcpyHostToDevice);

	//kernel call 
	const dim3 block(ntpb, ntpb);
	const dim3 grid(ceil((float)width / block.x), ceil((float)height / block.y));
	oilPaint << <grid, block >> >(d_src, d_dst, width, height);

	cudaDeviceSynchronize();

	Mat dst = Mat::zeros(src.size(), src.type());
	
	//copy back to host
	cudaMemcpy(dst.ptr(), d_dst, height * width * sizeof(uchar3), cudaMemcpyDeviceToHost);

	//free memory
	cudaFree(d_src);
	cudaFree(d_dst);

	imshow("Destination image", dst);

	//save oil painting version 
	std::vector<int> compression_params;
	compression_params.push_back(IMWRITE_PNG_COMPRESSION);
	compression_params.push_back(9);
	imwrite("oilVersion-" + fileName, dst, compression_params);

	waitKey();
	return 0;
}
